// Java3D packages
import java.applet.Applet;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.math.*;
import javax.media.j3d.*;
import javax.vecmath.*;

// Sun's utility packages for Java3D
import com.sun.j3d.loaders.objectfile.ObjectFile;
import com.sun.j3d.loaders.ParsingErrorException;
import com.sun.j3d.loaders.IncorrectFormatException;
import com.sun.j3d.loaders.Scene;
import com.sun.j3d.loaders.*;
import com.sun.j3d.utils.behaviors.mouse.*;
import com.sun.j3d.utils.behaviors.keyboard.*;
import com.sun.j3d.utils.picking.*;
import com.sun.j3d.utils.universe.*;

// The VRML97 loader package
import com.sun.j3d.loaders.vrml97.VrmlLoader;

// The joystick driver for Java3D on Linux
import net.dixonpeugh.j3d.linux.*;

/**
 * This class forms the center of the virtual robot. It is used to gain access
 * to all the important features of the simulator.
 *
 * @author Dieter Van Uytvanck
 * @version 0.1
 */
public class Cyclop extends Applet {
    
    private boolean noTriangulate = false;
    private boolean noStripify = false;
    
    // the number of robot parts
    private int nrParts = 7;
    
    // create an array holding the moving parts of the robot
    private BranchGroup [] parts = new BranchGroup[nrParts];
    private TransformGroup [] transformations = new TransformGroup[nrParts];
    private static Canvas3D mainCanvas3D, secondCanvas3D;
    private Bounds theBounds;
    
    // store the different angles of the robotparts
    private double baseAngle;
    private double shoulderAngle;
    private double elbowAngle;
    private double wristTiltAngle;
    private double wristTwistAngle;
    
    // position of the robot after initialization
    private static final double homeBaseAngle = 0.073;
    private static final double homeShoulderAngle = -2.07694;
    private static final double homeElbowAngle = 0.261799;
    private static final double homeWristTiltAngle = 2.00713;
    private static final double homeWristTwistAngle = 0.2;
    
    // maximum deviation of homeposition
    private static double allowedBaseAngle;
    private static double allowedShoulderAngleMin;
    private static double allowedShoulderAngleMax;
    private static double allowedElbowAngle;
    private static double allowedWristTiltAngle;
    private static double allowedWristTwistAngle;
    
    // the Transform3D's for all rotations needed
    private Transform3D baseRotation = new Transform3D();
    private Transform3D shoulderRotation = new Transform3D();
    private Transform3D elbowRotation = new Transform3D();
    private Transform3D wristTiltRotation = new Transform3D();
    private Transform3D wristTwistRotation = new Transform3D();
    private Transform3D viewRotation = new Transform3D();
    
    // is this a compensating move?
    private boolean compensateElbow = true;
    private boolean compensateWrist = true;
    
    private static BranchGroup theScene = new BranchGroup();
    private static CyclopGui theGui;
    private int sensitivity = 1;
    private static boolean simulation = false;
    private static CyclopCommunicationSender theCommunicationClient;
    // the different views on the robot
    private CyclopView mainView, secondView;
    // the array with the angles to send to the robot
    private double [] thePosition = new double [6];
    // send the input from the model to the real robot?
    private boolean sendInput = false;
    // open the socket to the robot?
    private boolean firstTime = true;
    private static JoystickInputDevice joyDevice;
    // the camera angles
    private double cameraX, cameraY, cameraZ, cameraZoom;
    // time stamp
    private long lastSent;
    
    
    /**
     * Creates an object that holds all the vital information for the simulation, 
     * for example the angles of the different parts of the robot.
     */
    
    public Cyclop() {
        
        setLayout(new BorderLayout());
        GraphicsConfiguration config =
        SimpleUniverse.getPreferredConfiguration();
        
        mainCanvas3D = new Canvas3D(config);
        secondCanvas3D = new Canvas3D(config);
        
        mainView = new CyclopView(mainCanvas3D);
        secondView = new CyclopView(secondCanvas3D);
        
        VirtualUniverse myUniverse = new VirtualUniverse();
        Locale myLocale = new Locale(myUniverse);
        BranchGroup myBranchGroup = new BranchGroup();
        
	myBranchGroup.addChild(mainView.getTransformGroup());
        myBranchGroup.addChild(secondView.getTransformGroup());

	myBranchGroup.compile();
	myLocale.addBranchGraph(myBranchGroup);                
        theScene = createSceneGraph();
        
        setMouseRotationBehavior();
        setJoystickRotationBehavior();
	setKeyBehavior();
        
	theScene.compile();
        
        myLocale.addBranchGraph(theScene);
    }


    /**
     * Creates the actual SceneGraph were all the objects are placed in.
     *
     * @return a BranchGroup containing the simulation graphics
     */
    private BranchGroup createSceneGraph() {
        // Create the root of the branch graph
        BranchGroup objRoot = new BranchGroup();
        
        // Create a Transformgroup to scale all objects so they
        // appear in the scene.
        TransformGroup objScale = new TransformGroup();
        Transform3D t3d = new Transform3D();
        t3d.setScale(0.8);
        objScale.setTransform(t3d);
        objRoot.addChild(objScale);
        
        // Create the transform group node and initialize it to the
        // identity.  Enable the TRANSFORM_WRITE capability so that
        // our behavior code can modify it at runtime.  Add it to the
        // root of the subgraph.
        
        for (int i=1; i<nrParts; i++) {
            transformations[i] = new TransformGroup();
            transformations[i].setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
            transformations[i].setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
            transformations[i].setCapability(TransformGroup.ENABLE_PICK_REPORTING);
            if (i==1)
                // top TransformGroup
                objScale.addChild(transformations[i]);
            else
                // child of previous TransformGroup
                transformations[i-1].addChild(transformations[i]) ;
        }
        
        int flags = ObjectFile.RESIZE;
        if (!noTriangulate) flags |= ObjectFile.TRIANGULATE;
        if (!noStripify) flags |= ObjectFile.STRIPIFY;
        
        for (int i=1; i<nrParts; i++) {
            Scene s = loadFile(i);
            parts[i]= s.getSceneGroup();
            transformations[i].addChild(parts[i]);
        }
        
        theBounds = new BoundingSphere(new Point3d(0.0,0.0,0.0), 100.0);
	setSceneProperties(objRoot);
        return objRoot;
    }
    
    
    
    
    /**
     * Loads the necessary VRML-file (part of the robot)
     * @param number the numer of the robot part
     * @return a Scene containing the robot part 
     */

    private Scene loadFile(int number) {
        Loader ldr = new VrmlLoader(Loader.LOAD_ALL);
        ldr.setFlags(Loader.LOAD_BEHAVIOR_NODES);
        Scene s = null;
        try {
            s = ldr.load("part" + number + ".wrl");
            return s;
        }
        catch (FileNotFoundException e) {
            System.err.println(e);
            System.exit(1);
        }
        catch (ParsingErrorException e) {
            System.err.println(e);
            System.exit(1);
        }
        catch (IncorrectFormatException e) {
            System.err.println(e);
            System.exit(1);
        }
        return null;
    }
    
    
    /** 
     * Rotates the base part of the robot 
     * @param angle the rotation angle in radians
     */
    private void baseRotate(double angle) {
        // verander de effectieve hoek
        baseAngle += angle;
        baseRotation.rotY(baseAngle);
        transformations[2].setTransform(baseRotation);
    }
    
    /** 
     * Rotates the shoulder part of the robot 
     * @param angle the rotation angle in radians
     */
    private void shoulderRotate(double angle) {
        double xtrans, ytrans, ztrans;
	// values from part3.vrml
        xtrans = 0.05; ytrans = 0.35; ztrans = 0;
        
        shoulderAngle += angle;
        
        // translation to the origin
        Transform3D trans = new Transform3D();
        trans.setTranslation(new Vector3d(-xtrans,-ytrans,-ztrans));
        
        // perform rotation
        shoulderRotation.rotZ(-shoulderAngle);
        shoulderRotation.mul(trans);
        
        // define translation away from the origin
        Transform3D backtrans = new Transform3D();
        backtrans.setTranslation(new Vector3d(xtrans,ytrans,ztrans));
        
	// perform return translation
        backtrans.mul(shoulderRotation);
        
        // make changes visible
        transformations[3].setTransform(backtrans);
        
        if (simulation && checkAngles(shoulderAngle, angle, wristTiltAngle)) {
            // compensate movements for dependent parts
            compensateWrist = false;
            elbowRotate(-angle);
            theGui.setElbowLabel(elbowAngle);
        }
    }
    

    /** 
     * Rotates the elbow part of the robot 
     * @param angle the rotation angle in radians
     */
    private void elbowRotate(double angle) {
	// comments: see shoulderRotate(angle)
        double xtrans, ytrans, ztrans;
        xtrans = 0.28; ytrans = 0.35; ztrans = 0;
        
        elbowAngle += angle;
        
        Transform3D trans = new Transform3D();
        trans.setTranslation(new Vector3d(-xtrans,-ytrans,-ztrans));
        
        elbowRotation.rotZ(-elbowAngle);
        elbowRotation.mul(trans);
        
        Transform3D temp = new Transform3D();
        temp.setTranslation(new Vector3d(xtrans,ytrans,ztrans));
        
        temp.mul(elbowRotation);
        transformations[4].setTransform(temp);
        
        if (simulation) {
            if (compensateWrist && checkAngles(shoulderAngle, angle, wristTiltAngle)) {
                // compensate movements
                wristTilt(-angle);
                theGui.setWristTiltLabel(wristTiltAngle);
            }
            else
		// don't compensate if this is a 'recursive compensation'
                compensateWrist = true;
        }
        
    }
    
    

    /** 
     * Tilts the wrist part of the robot 
     * @param angle the rotation angle in radians
     */
    public void wristTilt(double angle) {
	double xtrans, ytrans, ztrans;
        xtrans = 0.505; ytrans = 0.35; ztrans = 0;
        
        wristTiltAngle += angle;
        
	Transform3D trans = new Transform3D();
        trans.setTranslation(new Vector3d(-xtrans,-ytrans,-ztrans));
        
	wristTiltRotation.rotZ(-wristTiltAngle);
        wristTiltRotation.mul(trans);
        
	Transform3D backtrans = new Transform3D();
        backtrans.setTranslation(new Vector3d(xtrans,ytrans,ztrans));
        
	backtrans.mul(wristTiltRotation);
        transformations[5].setTransform(backtrans);
    }
    
    
    /** 
     * Twists the wrist part of the robot 
     * @param angle the rotation angle in radians
     */
    public void wristTwist(double angle) {
        
        double xtrans, ytrans, ztrans;
        xtrans = 0.505; ytrans = 0.37; ztrans = 0;
        
        wristTwistAngle += angle;
        
        Transform3D trans = new Transform3D();
        trans.setTranslation(new Vector3d(-xtrans,-ytrans,-ztrans));
        
        wristTwistRotation.rotY(wristTwistAngle);
        wristTwistRotation.mul(trans);
        
        Transform3D backtrans = new Transform3D();
        backtrans.setTranslation(new Vector3d(xtrans,ytrans,ztrans));
        
        backtrans.mul(wristTwistRotation);
        transformations[6].setTransform(backtrans);
    }
    
    
    /**
     * Starts the robot simulation.
     * @param args not used
     */
    public static void main(String[] args) {
        final Cyclop theCyclop = new Cyclop();
        theGui = new CyclopGui(theCyclop);
        theCyclop.initialise();
        CyclopCommunicationReceiver theCommunicationServer =
	    new CyclopCommunicationReceiver(theCyclop);
        theGui.setSize(800,600);
        theGui.show();
                
    }
    
    /**
     * Adds some MouseBehaviors and lights.
     * @param objRoot the BranchGroup containing the main Scene  
     */
    private void setSceneProperties(BranchGroup objRoot) {
        // Create the zoom behavior node
        MouseZoom behavior2 = new MouseZoom(MouseBehavior.INVERT_INPUT);
        behavior2.setTransformGroup(mainView.getTransformGroup());
	parts[1].addChild(behavior2);
	//theScene.addChild(behavior2);
        behavior2.setSchedulingBounds(theBounds);
        
        // Create the translate behavior node
        MouseTranslate behavior3 = new MouseTranslate(MouseBehavior.INVERT_INPUT);
        //BEHAVIOR3.setTransformGroup(transformations[1]);
	behavior3.setTransformGroup(mainView.getTransformGroup());
        parts[1].addChild(behavior3);
        behavior3.setSchedulingBounds(theBounds);
        
	// Set up the background
        Color3f bgColor = new Color3f(0.05f, 0.05f, 0.5f);
        Background bgNode = new Background(bgColor);
        bgNode.setApplicationBounds(theBounds);
 objRoot.addChild(bgNode);
        
        // Set up the ambient light
        Color3f ambientColor = new Color3f(0.1f, 0.1f, 0.1f);
        AmbientLight ambientLightNode = new AmbientLight(ambientColor);
        ambientLightNode.setInfluencingBounds(theBounds);
        objRoot.addChild(ambientLightNode);
        
        // Set up the directional lights
        Color3f light1Color = new Color3f(1.0f, 1.0f, 0.9f);
        Vector3f light1Direction  = new Vector3f(4.0f, -7.0f, -12.0f);
        Color3f light2Color = new Color3f(0.3f, 0.3f, 0.4f);
        Vector3f light2Direction  = new Vector3f(-6.0f, -2.0f, -1.0f);
        
        DirectionalLight light1
        = new DirectionalLight(light1Color, light1Direction);
        light1.setInfluencingBounds(theBounds);
        objRoot.addChild(light1);
        
        DirectionalLight light2
        = new DirectionalLight(light2Color, light2Direction);
        light2.setInfluencingBounds(theBounds);
        objRoot.addChild(light2);
    }
    
    /**
     * Enables the rotation behaviors associated with the rotating parts of the
     * robot.
     */
    private void setMouseRotationBehavior() {
        
        // set rotation behavior for the Base part
       MouseRotateOneAxisBase dragBase = new MouseRotateOneAxisBase(MouseRotateOneAxis.MANUAL_WAKEUP, this);
        PickRotateBehaviorOneAxis baseRotateBehavior =
        new PickRotateBehaviorOneAxis(MouseRotateOneAxis.X_AXIS,
        parts[2], mainCanvas3D, theBounds, dragBase);
        transformations[2].addChild(baseRotateBehavior);
        
        // set rotation behavior for the shoulder part
        MouseRotateOneAxisShoulder dragShoulder =
        new MouseRotateOneAxisShoulder(MouseRotateOneAxis.MANUAL_WAKEUP, this);
        PickRotateBehaviorOneAxis shoulderRotateBehavior =
        new PickRotateBehaviorOneAxis(MouseRotateOneAxis.Y_AXIS,
        parts[3], mainCanvas3D, theBounds, dragShoulder);
        transformations[3].addChild(shoulderRotateBehavior);
        
        // set rotation behavior for the elbow part
        MouseRotateOneAxisElbow dragElbow =
        new MouseRotateOneAxisElbow(MouseRotateOneAxis.MANUAL_WAKEUP, this);
        PickRotateBehaviorOneAxis elbowRotateBehavior =
        new PickRotateBehaviorOneAxis(MouseRotateOneAxis.X_AXIS,
        parts[4], mainCanvas3D, theBounds, dragElbow);
        transformations[4].addChild(elbowRotateBehavior);
        
        // set rotation behavior for the wristTilt part
        MouseRotateOneAxisWristTilt dragWristTilt =
        new MouseRotateOneAxisWristTilt(MouseRotateOneAxis.MANUAL_WAKEUP, this);
        PickRotateBehaviorOneAxis wristTiltBehavior =
        new PickRotateBehaviorOneAxis(MouseRotateOneAxis.Y_AXIS,
        parts[5], mainCanvas3D, theBounds, dragWristTilt);
        transformations[5].addChild(wristTiltBehavior);
    }
    
    /**
     * Places the robot and the views in the "home"-position.
     */
    public void initialise() {
        
        baseAngle = 0;
        shoulderAngle = 0;
        elbowAngle = 0;
        wristTiltAngle = 0;
        wristTwistAngle = 0;
        
        // max- and min-values for the communication with the real robot
        allowedBaseAngle = homeBaseAngle + (150d/180d)*Math.PI;
        allowedShoulderAngleMin = homeShoulderAngle;
        allowedShoulderAngleMax = 0.5;
        allowedElbowAngle = homeElbowAngle + Math.PI/2;
        allowedWristTiltAngle = homeWristTiltAngle + Math.PI;
        allowedWristTwistAngle= homeWristTwistAngle + Math.PI;
        
        // start the robot with these angles
        baseRotate(homeBaseAngle);
        shoulderRotate(homeShoulderAngle);
        elbowRotate(homeElbowAngle);
        wristTilt(homeWristTiltAngle);
        wristTwist(homeWristTwistAngle);
        
        theGui.setBaseLabel(homeBaseAngle);
        theGui.setShoulderLabel(homeShoulderAngle);
        theGui.setElbowLabel(homeElbowAngle);
        theGui.setWristTiltLabel(homeWristTiltAngle);
        theGui.setWristTwistLabel(homeWristTwistAngle);
        
        cameraZoom = 2.41;
        setCameraX(0); 
        setCameraY(0.3);
        cameraZ = -cameraZoom;
        
	Transform3D cameraTransform = new Transform3D();   
        cameraTransform.lookAt( new Point3d(0, 1.7, 0),
				new Point3d(0,0,0),
				new Vector3d(0,0,-1));
        cameraTransform.invert();
        secondView.getTransformGroup().setTransform(cameraTransform);
        
    }
    
    
    /**
     * Closes all open resources (for example input devices and sockets).
     */
    public static void exit() {
        try {
            theCommunicationClient.closeSocket();
            closeJoystick();
        }
        catch (Exception e) {
	    System.err.println("Error while closing resources: " + e);
	}
    }
    
    
    /**
     * Rotates the base of the robot after checking some constraints.
     * @param angle The rotation angle.
     * @param isBehavior Indicates whether the input is generated by a behavior
     * or manual input. If <code>true</code> then a delay threshold is applied when 
     * sending data to the real robot. Otherwise the robot would get too much 
     * information, which results in very slow movements because of the amount of
     * interpolations needed.
     */
    public void checkBaseRotate(double angle, boolean isBehavior) {
        double adjustAngle = angle/sensitivity;
        double futureAngle = baseAngle + adjustAngle;
        
        // simulation mode, check robot movements
        if (simulation) {
            if (futureAngle <= allowedBaseAngle && futureAngle >= -allowedBaseAngle) {
                baseRotate(adjustAngle);
                theGui.setBaseLabel(futureAngle);
                if (sendInput)
                    sendToRobot(futureAngle, shoulderAngle, elbowAngle, 
				wristTiltAngle, wristTwistAngle, 0, isBehavior);
            }
        }
        // do not check movements
        else {
            baseRotate(adjustAngle);
            theGui.setBaseLabel(futureAngle);
            if (sendInput)
                sendToRobot(futureAngle, shoulderAngle, elbowAngle, 
			    wristTiltAngle, wristTwistAngle, 0, isBehavior);
        }
        
    }
    

    /**
     * Rotates the shoulder of the robot after checking some constraints.
     * @param angle The rotation angle.
     * @param isBehavior Indicates whether the input is generated by a behavior
     * or manual input. If <code>true</code> then a delay threshold is applied when 
     * sending data to the real robot. Otherwise the robot would get too much 
     * information, which results in very slow movements because of the amount of
     * interpolations needed.
     */
    public void checkShoulderRotate(double angle, boolean isBehavior) {
        double adjustAngle = angle/sensitivity;
        double futureAngle = shoulderAngle + adjustAngle;
        
        if (simulation) {
            if (futureAngle <= allowedShoulderAngleMax && futureAngle >= allowedShoulderAngleMin && 
                 checkAngles(futureAngle, elbowAngle, wristTiltAngle)) {
                shoulderRotate(adjustAngle);
                theGui.setShoulderLabel(futureAngle);
                if (sendInput)
                    sendToRobot(baseAngle, futureAngle, elbowAngle, 
				wristTiltAngle, wristTwistAngle, 0, isBehavior);
            }
        }
        else {
            shoulderRotate(adjustAngle);
            theGui.setShoulderLabel(futureAngle);
            if (sendInput)
                sendToRobot(baseAngle, futureAngle, elbowAngle, wristTiltAngle, 
			    wristTwistAngle, 0, isBehavior);
        }
        
    }
    
    
    /**
     * Rotates the elbow of the robot after checking some constraints.
     * @param angle The rotation angle.
     * @param isBehavior Indicates whether the input is generated by a behavior
     * or manual input. If <code>true</code> then a delay threshold is applied when 
     * sending data to the real robot. Otherwise the robot would get too much 
     * information, which results in very slow movements because of the amount of
     * interpolations needed.
     */
    public void checkElbowRotate(double angle, boolean isBehavior) {
        
        double adjustAngle = angle/sensitivity;
        double futureAngle = elbowAngle + adjustAngle;
        
        if (simulation) {
            if (//futureAngle <= allowedElbowAngle && futureAngle >= -allowedElbowAngle &&
                checkAngles(shoulderAngle, futureAngle, wristTiltAngle)) {
                elbowRotate(adjustAngle);
                theGui.setElbowLabel(futureAngle);
                if (sendInput)
                    sendToRobot(baseAngle, shoulderAngle, futureAngle, 
				wristTiltAngle, wristTwistAngle, 0, isBehavior);
            }
        }
        else {
            elbowRotate(angle/sensitivity);
            theGui.setElbowLabel(futureAngle);
            if (sendInput)
                sendToRobot(baseAngle, shoulderAngle, futureAngle, 
			    wristTiltAngle, wristTwistAngle, 0, isBehavior);
        }
    }
    
    
    /**
     * Tilts the wrist of the robot after checking some constraints.
     * @param angle The rotation angle.
     * @param isBehavior Indicates whether the input is generated by a behavior
     * or manual input. If <code>true</code> then a delay threshold is applied when 
     * sending data to the real robot. Otherwise the robot would get too much 
     * information, which results in very slow movements because of the amount of
     * interpolations needed.
     */
    public void checkWristTilt(double angle, boolean isBehavior) {
        double adjustAngle = angle/sensitivity;
        double futureAngle = wristTiltAngle + adjustAngle;
        
        if (simulation) {
            if (futureAngle <= allowedWristTiltAngle && futureAngle >= -allowedWristTiltAngle &&
            checkAngles(shoulderAngle, elbowAngle, futureAngle)) {
                wristTilt(adjustAngle);
                theGui.setWristTiltLabel(futureAngle);
                if (sendInput)
                    sendToRobot(baseAngle, shoulderAngle, elbowAngle, 
				futureAngle, wristTwistAngle, 0, isBehavior);
            }
        }
        else {
            wristTilt(adjustAngle);
            theGui.setWristTiltLabel(futureAngle);
            if (sendInput)
                sendToRobot(baseAngle, shoulderAngle, elbowAngle, 
			    futureAngle, wristTwistAngle, 0, isBehavior);
        }
    }
    
    
    /**
     * Twists the wrist of the robot after checking some constraints.
     * @param angle The rotation angle.
     * @param isBehavior Indicates whether the input is generated by a behavior
     * or manual input. If <code>true</code> then a delay threshold is applied when 
     * sending data to the real robot. Otherwise the robot would get too much 
     * information, which results in very slow movements because of the amount of
     * interpolations needed.
     */
    public void checkWristTwist(double angle, boolean isBehavior) {
        double adjustAngle = angle/sensitivity;
        double futureAngle = wristTwistAngle + adjustAngle;
        
        if (simulation) {
            if (futureAngle <= allowedWristTwistAngle && futureAngle >= -allowedWristTwistAngle) {
                wristTwist(adjustAngle);
                theGui.setWristTwistLabel(futureAngle);
                if (sendInput)
                    sendToRobot(baseAngle, shoulderAngle, elbowAngle, 
				wristTiltAngle, futureAngle, 0, isBehavior);
            }
        }
        else {
            wristTwist(adjustAngle);
            theGui.setWristTwistLabel(futureAngle);
            if (sendInput)
                sendToRobot(baseAngle, shoulderAngle, elbowAngle, 
			    wristTiltAngle, futureAngle, 0, isBehavior);
        }
    }
    
    
    /**
     * Returns the Canvas3D used for rendering the main view
     * @return the main Canvas3D
     */
    public Canvas3D getMainCanvas() {
        return mainCanvas3D;
    }
    
    /**
     * Returns the Canvas3D used for rendering the second view. It could contain
     * a top view, for example.
     * @return the second Canvas3D 
     */
    public Canvas3D getSecondCanvas() {
        return secondCanvas3D;
    }
    
    /** 
     * Returns the BranchGraph which contains the whole Scene 
     * @return the whole Scene
     */
    public BranchGroup getTheScene() {
        return theScene;
    }
    
    /**
     * Returns the sensitivity used for input on a scale from 1 (not very 
     * sensitive) to 10 (more sensitive).
     * @return a sensitivity degree
     */
    public int getSensitivity() {
        return sensitivity;
    }
    
    /**
     * Sets the sensitivity degree between 1 and 10.
     * @param input the sensitivity degree
     */
    public void setSensitivity(int input) {
        if (1 <= input && input <= 10)
	    sensitivity = input;
    }
    
    /**
     * Attaches the joystick's rotation behavior to the Scene Graph.
     */
    public void setJoystickRotationBehavior() {
	joyDevice = new JoystickInputDevice();
        joyDevice.initialize();
        joyDevice.setProcessingMode(InputDevice.DEMAND_DRIVEN);
	mainView.getView().getPhysicalEnvironment().addInputDevice(joyDevice);
	Sensor joySensor = joyDevice.getSensor(0);
	// set the joystick behavior
        JoystickBehavior rotBehavior = new JoystickBehavior(joySensor, this);
        rotBehavior.setSchedulingBounds 
	    (new BoundingSphere(new Point3d(0.0,0.0,0.0), Float.MAX_VALUE));
        theScene.addChild(rotBehavior);
    }
    

    /**
     * Enables a walk-through behavior for the scene.
     */
    // not used because of a bug in Blackdown's Java3D v. 1.3 beta
    private void setKeyBehavior() {
	/* KeyNavigatorBehavior keyNav = new KeyNavigatorBehavior(mainView.getTransformGroup());
	   keyNav.setSchedulingBounds(theBounds);
	   theScene.addChild(keyNav); */
    }	

    /** Closes the joystick input device */
    public static void closeJoystick() {
        joyDevice.close();
    }
    
    /** 
     * Enables the simulation mode for input. This makes the virtual robot check
     * the input and enables "compensating moves". This means that the movement
     * of one part does not influence the parts above that part.
     * @param simulation If <code>true</code> simulation mode is enabled. 
    */
    public void setSimulationMode(boolean simulation) {
        this.simulation = simulation;
    }
    
    /**
     * Returns whether the virtual robot runs in simulation mode.
     * @return the simulation state
     */
    public boolean getSimulationMode() {
        return simulation;
    }
    
    /**
     * Enables/disbles the transmission of the input to the real robot.
     * @param mode send the input to real robot
     */
    public void setSendinputMode(boolean mode) {
        this.sendInput = mode;
        if (mode == true && firstTime) {
            theCommunicationClient = new CyclopCommunicationSender();
            firstTime = false;
        }
    }
    
    /**
     * Returns if the model transfers input to real robot.
     * @return the current SendInput mode
     */
    public boolean getSendinputMode() {
        return sendInput;
    }
    
    
    /** 
     * Performs a rotation of the base and do not propagate result to the real robot.
     * @param angle the rotation angle
     */
    public void setBaseAngle(double angle) {
        if (!sendInput) {
            double difference = angle - baseAngle;
            baseRotate(difference);
            theGui.setBaseLabel(angle);
        }
    }
    

    /** 
     * Performs a rotation of the shoulder and do not propagate result to the real robot.
     * @param angle the rotation angle
     */   
    public void setShoulderAngle(double angle) {
        if (!sendInput) {
            double difference = angle - shoulderAngle;
            shoulderRotate(difference);
            theGui.setShoulderLabel(angle);
        }
    }
    

    /** 
     * Performs a rotation of the elbow and do not propagate result to the real robot.
     * @param angle the rotation angle
     */
    public void setElbowAngle(double angle) {
        if (!sendInput) {
            double difference = angle - elbowAngle;
            elbowRotate(difference);
            theGui.setElbowLabel(angle);
        }
    }
    
    /** 
     * Performs a tilt of the wrist and do not propagate result to the real robot.
     * @param angle the rotation angle
     */
    public void setWristTiltAngle(double angle) {
        if (!sendInput) {
            double difference = angle - wristTiltAngle;
            wristTilt(difference);
            theGui.setWristTiltLabel(angle);
        }
    }
    
    /** 
     * Performs a twist of the wrist and do not propagate result to the real robot.
     * @param angle the rotation angle
     */
    public void setWristTwistAngle(double angle) {
        if (!sendInput) {
            double difference = angle - wristTwistAngle;
            wristTwist(difference);
            theGui.setWristTwistLabel(angle);
        }
    }
    
    /**
     * Sends the joint coordinates to the real robot.
     * @param b BaseAngle
     * @param s ShoulderAngle
     * @param e ElbowAngle
     * @param wtw WristTwistAngle
     * @param wti WristTiltAngle
     * @param isBehavior behavior mode
     */
    private void sendToRobot(double b, double s, double e, double wti, double wtw, 
			     double g, boolean isBehavior) {
    	// if the input comes from a behavior, respect a threshold of 100 ms
	if (isBehavior && System.currentTimeMillis() - lastSent < 100) 
	    return;	
	
        thePosition[0] = b;
        thePosition[1] = s;
        thePosition[2] = e;
        thePosition[3] = wti;
        thePosition[4] = wtw;
        thePosition[5] = g;
        try {
            theCommunicationClient.send(thePosition);
        } catch (NullPointerException ex) {
            System.err.println("No communcation possible with robot");
        }
	lastSent =  System.currentTimeMillis();
    }
    

    /**
     * Changes the main camera's view in a circular way in the XZ-plane around 
     * the Y-axis.
     * @param angle the rotation angle
     */
    public void setCameraX(double angle) {
        Transform3D temp = new Transform3D();
                
        if (Math.abs(angle) <= 1) {
            cameraX = cameraZoom * angle;
            cameraZ = Math.sqrt(cameraZoom*cameraZoom - cameraX*cameraX);
        }
        else {
            if (angle <-1) cameraX = -(2 - Math.abs(angle))*cameraZoom;
            else cameraX = (2 - Math.abs(angle))*cameraZoom;
            cameraZ = -Math.sqrt(cameraZoom*cameraZoom - cameraX*cameraX);
        }
        
        temp.lookAt(new Point3d(cameraX, cameraY, cameraZ),
                    new Point3d(0,0,0),
                    new Vector3d(0,1,0));
        
        temp.invert();
        mainView.getTransformGroup().setTransform(temp);
    }
    
      
    
    /**
     *  Makes the camera's current position sink or raise.
     * @param angle the rotation angle
     */
    public void setCameraY(double angle) {
        Transform3D temp = new Transform3D();
        cameraY = 2*angle;
        
        Vector3d viewDirection;
        viewDirection = new Vector3d(0,1,0);
        
        temp.lookAt( new Point3d(cameraX, cameraY, cameraZ),
        new Point3d(0,0.1,0),
        viewDirection);
        temp.invert();
        mainView.getTransformGroup().setTransform(temp);
    }
    

    /**
     * Changes the distance between the camera and the origin, so a sort of 
     * zoom function.
     * @param zoom the zoom factor
     */
    public void setCameraZ(double zoom) {
        cameraZoom = 2.41 - (zoom * 2.41) + 1;
    }

  
    /**
     * Checks wether a combination of angles is allowed.
     * @param s ShoulderAngle
     * @param e ElbowAngle
     * @param wt WristTwistAngle
     */
    private boolean checkAngles (double s, double e, double wt) {
        return ( Math.abs(s) < 2.6 &&  // minimum hoek van 30
                 Math.abs(e) < 2.6 &&
                 wt < 4d*Math.PI/3d && wt > -Math.PI/3d
                 && !( s > 0.2 && e < 0)); // hoek tss -60 en 240
        
    }
            
}



/*
 * Quote of the day :-) :
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */
