/* Author Fehlmann Andreas
 * Copyright (c) 2000  University of applied sciences
 * Biel School of Engineering and Architecture, Switzerland.
 * http://www.hta-bi.bfh.ch
 * All Rights Reserved.
 *
 * Project FurnishingPlanner is under the GNU General public License
 *
 *
 * This class is a modification of the MouseRotate class from Sun:
 *
 *	@(#)MouseRotate.java 1.25 00/04/11 17:31:19
 *
 * Copyright (c) 1996-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Sun grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to Sun.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */

import com.sun.j3d.utils.behaviors.mouse.MouseBehavior;
import com.sun.j3d.utils.behaviors.mouse.MouseBehaviorCallback;
import java.awt.AWTEvent;
import java.awt.event.MouseEvent;
import java.util.Enumeration;
import javax.media.j3d.*;
import javax.vecmath.*;



/**
 * MouseRotateOneAxis is a Java3D behavior object that lets users control the
 * rotation of an object via a mouse. The object can be rotated around the x, y or z axis.
 * <p>
 * To use this utility, first create a transform group that this
 * rotate behavior will operate on. Then,
 *<blockquote><pre>
 *
 *   MouseRotateOneAxis behavior = new MouseRotateOneAxis();
 *   behavior.setTransformGroup(objTrans);
 *   objTrans.addChild(behavior);
 *   behavior.setSchedulingBounds(bounds);
 *
 *</pre></blockquote>
 * The above code will add the rotate behavior to the transform
 * group. The user can rotate any object attached to the objTrans.
 */

public class MouseRotateOneAxis extends MouseBehavior {
    public static final int X_AXIS = 0;
    public static final int Y_AXIS = 1;
    public static final int Z_AXIS = 2;
    
    double x_angle, y_angle;
    double x_factor = .03;
    double y_factor = .03;
    int axisOfRotation = X_AXIS;
    Transform3D transformZ = new Transform3D();
    
    protected MouseBehaviorCallback callback = null;
    protected Cyclop refCyclop;
     
    
    /**
     * Creates a rotate behavior given the transform group.
     * @param transformGroup The transformGroup to operate on.
     */
    public MouseRotateOneAxis(TransformGroup transformGroup) {
        super(transformGroup);
    }
    
    
    /**
    * Creates a default rotate behavior.
     */
    public MouseRotateOneAxis() {
        super(0);
    }
    
    /**
     * Creates a rotate behavior.
     * Note that this behavior still needs a transform
     * group to work on (use setTransformGroup(tg)) and
     * the transform group must add this behavior.
     * @param flags interesting flags (wakeup conditions).
     */
    public MouseRotateOneAxis(int flags) {
	super(flags);
    }


    /**
     * Creates a rotate behavior.
     * Note that this behavior still needs a transform
     * group to work on (use setTransformGroup(tg)) and
     * the transform group must add this behavior.
     * @param flags interesting flags (wakeup conditions).
     * @param c reference to the Cylop object
     * @see Cyclop
     */
    public MouseRotateOneAxis(int flags, Cyclop c) {
        super(flags);
	refCyclop = c;
    }
    
        
    /** Initizalizes this behavior */
    public void initialize() {
        super.initialize();
        x_angle = 0;
        y_angle = 0;
        if ((flags & INVERT_INPUT) == INVERT_INPUT) {
            invert = true;
            x_factor *= -1;
            y_factor *= -1;
        }
    }
    
    
    /**
     * Returns the integer value of the current axis.
     * @return current rotation axis
     */
    public int getAxisOfRotation() {
        return axisOfRotation;
    }
    
    
    /**
     * Set the axis of rotation. Acceptable values are X_AXIS, Y_AXIS and Z_AXIS
     * @param axis rotation axis
     */
    public void setAxisOfRotation(int axis) {
        if ((axis == X_AXIS) || (axis == Y_AXIS) || (axis == Z_AXIS))
            axisOfRotation = axis;
    }
    
    
    /**
     * Return the x-axis movement multipler.
     */
    public double getXFactor() {
        return x_factor;
    }
    
    /**
     * Return the y-axis movement multipler.
     **/
    public double getYFactor() {
        return y_factor;
    }
    
    
    /**
     * Set the x-axis amd y-axis movement multipler with factor.
     */
    public void setFactor( double factor) {
        x_factor = y_factor = factor;
    }
    
    
    /**
     * Set the x-axis amd y-axis movement multipler with xFactor and yFactor
     * respectively.
     */
    public void setFactor( double xFactor, double yFactor) {
        x_factor = xFactor;
        y_factor = yFactor;
    }
    
    
    /**
     * This method is invoked by the behavior scheduler.
     * @param criteria An enumeration of the criteria that caused the stimulus.
     */
    public void processStimulus(Enumeration criteria) {
        

	WakeupCriterion wakeup;
        AWTEvent[] events;
        MouseEvent evt;
        int id;
        int dx, dy;
        
        while (criteria.hasMoreElements()) {
            wakeup = (WakeupCriterion) criteria.nextElement();
            if (wakeup instanceof WakeupOnAWTEvent) {
                events = ((WakeupOnAWTEvent)wakeup).getAWTEvent();
                if (events.length > 0) {
                    evt = (MouseEvent) events[events.length-1];
                    processMouseEvent(evt);
                    if (((buttonPress)&&((flags & MANUAL_WAKEUP) == 0)) ||
                    ((wakeUp)&&((flags & MANUAL_WAKEUP) != 0))) {
                        id = evt.getID();
                        if ((id == MouseEvent.MOUSE_DRAGGED) &&
                        !evt.isMetaDown() && ! evt.isAltDown()){
                            x = evt.getX();
                            y = evt.getY();
                            
                            dx = x - x_last;
                            dy = y - y_last;
                            
                            if (!reset){
                                x_angle = dy * y_factor;
                                y_angle = dx * x_factor;
                                
                                if (axisOfRotation == X_AXIS) {
                                    transformX.rotX(x_angle);
                                    transformY.rotY(0);
                                    transformZ.rotZ(0);
                                }
                                else if (axisOfRotation == Y_AXIS) {
                                    transformX.rotX(0);
                                    transformY.rotY(y_angle);
                                    transformZ.rotZ(0);
                                }
                                else if (axisOfRotation == Z_AXIS) {
                                    transformX.rotX(0);
                                    transformY.rotY(0);
                                    transformZ.rotZ(y_angle);
                                }
                                
                                transformGroup.getTransform(currXform);
                                
                                Matrix4d mat = new Matrix4d();
                                
                                // Remember old matrix
                                currXform.get(mat);
                                
                                // Translate to origin
                                currXform.setTranslation(new Vector3d(0.0,0.0,0.0));
                                if (invert) {
                                    currXform.mul(currXform, transformX);
                                    currXform.mul(currXform, transformY);
                                    currXform.mul(currXform, transformZ);
                                } else {
                                    currXform.mul(transformX, currXform);
                                    currXform.mul(transformY, currXform);
                                    currXform.mul(transformZ, currXform);
                                }
                                
                                // Set old translation back
                                Vector3d translation = new
                                Vector3d(mat.m03, mat.m13, mat.m23);
                                currXform.setTranslation(translation);
                                
                                // Update xform
                                transformGroup.setTransform(currXform);
                                
                                transformChanged(currXform );
                                
                                if (callback!=null)
                                    callback.transformChanged(MouseBehaviorCallback.ROTATE,
                                    currXform );
                            }
                            else {
                                reset = false;
                            }
                            
                            x_last = x;
                            y_last = y;
                        }
                        else if (id == MouseEvent.MOUSE_PRESSED) {
                            x_last = evt.getX();
                            y_last = evt.getY();
                        }
                    }
                }
            }
        }
        
        wakeupOn(mouseCriterion);
    }
    
    
    /**
     * Users can overload this method  which is called every time
     * the Behavior updates the transform
     *
     * Default implementation does nothing
     */
    public void transformChanged( Transform3D transform ) {
    }
    
    
    /**
     * The transformChanged method in the callback class will
     * be called every time the transform is updated
     */
    public void setupCallback( MouseBehaviorCallback callback ) {
        this.callback = callback;
    }
}
